"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports["default"] = void 0;

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _LayerUpdateState = _interopRequireDefault(require("./LayerUpdateState"));

var _ObjectRemovalHelper = _interopRequireDefault(require("../Process/ObjectRemovalHelper"));

var _Layer2 = _interopRequireDefault(require("./Layer"));

var _Coordinates = _interopRequireDefault(require("../Core/Geographic/Coordinates"));

var _Extent = _interopRequireDefault(require("../Core/Geographic/Extent"));

var _Label = _interopRequireDefault(require("../Core/Label"));

var _Feature = require("../Core/Feature");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function () {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function () { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

var coord = new _Coordinates["default"]('EPSG:4326', 0, 0, 0);
var content;

var _extent = new _Extent["default"]('EPSG:4326', 0, 0, 0, 0);
/**
 * A layer to handle a bunch of `Label`. This layer can be created on its own,
 * but it is better to use the option `labelEnabled` on another `Layer` to let
 * it work with it (see the `vector_tile_raster_2d` example).
 *
 * @property {boolean} isLabelLayer - Used to checkout whether this layer is a
 * LabelLayer.  Default is true. You should not change this, as it is used
 * internally for optimisation.
 * @property {string} crs - The crs of this layer.
 */


var LabelLayer = /*#__PURE__*/function (_Layer) {
  (0, _inherits2["default"])(LabelLayer, _Layer);

  var _super = _createSuper(LabelLayer);

  function LabelLayer(id, crs) {
    var _this;

    (0, _classCallCheck2["default"])(this, LabelLayer);
    _this = _super.call(this, id);
    _this.isLabelLayer = true;
    _this.crs = crs;
    _this.domElement = document.createElement('div');

    _this.defineLayerProperty('visible', true, function () {
      _this.domElement.style.display = _this.visible ? 'block' : 'none';
    });

    return _this;
  }
  /**
   * Reads each {@link FeatureGeometry} that contains label configuration, and
   * creates the corresponding {@link Label}. To create a `Label`, a geometry
   * needs to have a `label` object with at least a few properties:
   * - `content`, which refers to `Label#content`
   * - `position`, which refers to `Label#position`
   * - (optional) `config`, containing miscellaneous configuration for the
   *   label
   *
   * The geometry (or its parent Feature) needs to have a Style set.
   *
   * @param {FeatureCollection} data - The FeatureCollection to read the
   * labels from.
   * @param {Extent} extent
   *
   * @return {Label[]} An array containing all the created labels.
   */


  (0, _createClass2["default"])(LabelLayer, [{
    key: "convert",
    value: function convert(data, extent) {
      var _this2 = this;

      var labels = [];
      var layerField = this.style && this.style.text && this.style.text.field; // Converting the extent now is faster for further operation

      extent.as(data.crs, _extent);
      coord.crs = data.crs;
      data.features.forEach(function (f) {
        // TODO: add support for LINE and POLYGON
        if (f.type !== _Feature.FEATURE_TYPES.POINT) {
          return;
        }

        var featureField = f.style && f.style.text.field;
        f.geometries.forEach(function (g) {
          var minzoom = g.properties.style && g.properties.style.zoom.min || f.style && f.style.zoom.min || _this2.style && _this2.style.zoom && _this2.style.zoom.min; // Don't create a label if it is in-between two steps of zoom

          if (!_this2.source.isFileSource) {
            if (data.extent.zoom != minzoom) {
              return;
            }
          } else if (extent.zoom != minzoom) {
            return;
          } // NOTE: this only works because only POINT is supported, it
          // needs more work for LINE and POLYGON


          coord.setFromArray(f.vertices, g.size * g.indices[0].offset);
          data.transformCoordinates(coord);

          if (f.size == 2) {
            coord.z = 0;
          }

          if (!_extent.isPointInside(coord)) {
            return;
          }

          var geometryField = g.properties.style && g.properties.style.text.field;

          if (!geometryField && !featureField && !layerField) {
            return;
          } else if (geometryField) {
            content = g.properties.style.getTextFromProperties(g.properties);
          } else if (featureField) {
            content = f.style.getTextFromProperties(g.properties);
          } else if (layerField) {
            content = _this2.style.getTextFromProperties(g.properties);
          }

          var label = new _Label["default"](content, coord.clone(), g.properties.style || f.style || _this2.style);

          if (f.size == 2) {
            label.needsAltitude = true;
          }

          labels.push(label);
        });
      });
      return labels;
    } // placeholder

  }, {
    key: "preUpdate",
    value: function preUpdate() {}
  }, {
    key: "update",
    value: function update(context, layer, node, parent) {
      var _this3 = this;

      if (!parent && node.children.length) {
        // if node has been removed dispose three.js resource
        _ObjectRemovalHelper["default"].removeChildrenAndCleanupRecursively(this, node);

        return;
      }

      if (this.frozen || !node.visible || !this.visible) {
        return;
      }

      if (node.layerUpdateState[this.id] === undefined) {
        node.layerUpdateState[this.id] = new _LayerUpdateState["default"]();
      }

      var elevationLayer = node.material.getElevationLayer();

      if (elevationLayer && node.layerUpdateState[elevationLayer.id].canTryUpdate()) {
        node.children.forEach(function (c) {
          if (c.isLabel && c.needsAltitude && c.updateElevationFromLayer(_this3.parent)) {
            c.update3dPosition(_this3.crs);
          }
        });
      }

      if (!node.layerUpdateState[this.id].canTryUpdate()) {
        return;
      }

      var extentsDestination = node.getExtentsByProjection(this.source.projection) || [node.extent];
      var extentsSource = [];

      var _iterator = _createForOfIteratorHelper(extentsDestination),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          var extentDest = _step.value;
          var ext = this.source.projection == extentDest.crs ? extentDest : extentDest.as(this.source.projection);

          if (!this.source.extentInsideLimit(ext)) {
            node.layerUpdateState[this.id].noMoreUpdatePossible();
            return;
          }

          extentsSource.push(extentDest);
        }
      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }

      node.layerUpdateState[this.id].newTry();
      var command = {
        layer: this,
        extentsSource: extentsSource,
        view: context.view,
        threejsLayer: this.threejsLayer,
        requester: node
      };
      return context.scheduler.execute(command).then(function (result) {
        if (!result) {
          return;
        }

        var renderer = context.view.mainLoop.gfxEngine.label2dRenderer;
        var labelsDiv = [];
        result.forEach(function (labels) {
          if (!node.parent) {
            labels.forEach(function (l) {
              _ObjectRemovalHelper["default"].removeChildrenAndCleanupRecursively(_this3, l);

              renderer.removeLabelDOM(l);
            });
            return;
          }

          labels.forEach(function (label) {
            if (label.needsAltitude) {
              label.updateElevationFromLayer(_this3.parent);
            }

            node.add(label);
            label.update3dPosition(_this3.crs);
            labelsDiv.push(label.content);
          });
        });

        if (labelsDiv.length > 0) {
          var _node$domElement;

          // Add all labels for this tile at once to batch it
          node.domElement = document.createElement('div');

          (_node$domElement = node.domElement).append.apply(_node$domElement, labelsDiv);

          (node.findClosestDomElement() || _this3.domElement).appendChild(node.domElement);

          node.domElementVisible = true; // Batch update the dimensions of labels all at once to avoid
          // redraw for at least this tile.

          result.forEach(function (labels) {
            return labels.forEach(function (label) {
              return label.initDimensions();
            });
          });
          result.forEach(function (labels) {
            return labels.forEach(function (label) {
              label.visible = false;
            });
          }); // Sort labels so they can be the first in the renderer. That
          // way, we cull labels on parent tile first, and then on
          // children tile. This allows a z-order priority, and reduce
          // flickering.

          node.children.sort(function (c) {
            return c.isLabel ? -1 : 1;
          }); // Necessary event listener, to remove any Label attached to
          // this tile

          node.addEventListener('removed', function () {
            renderer.hideNodeDOM(node);
            result.forEach(function (labels) {
              return labels.forEach(function (label) {
                return node.remove(label);
              });
            });
            node.domElement.parentElement.removeChild(node.domElement);
          });
        }

        node.layerUpdateState[_this3.id].noMoreUpdatePossible();
      });
    }
  }]);
  return LabelLayer;
}(_Layer2["default"]);

var _default = LabelLayer;
exports["default"] = _default;